/*
 * dotmatrx.c --- dot matrix functions
 * 
 * Copyright (c) 2019 Mariano Alvarez Fernandez (malfer@telefonica.net)
 *
 * This file is part of FEdit, a visual font editor for the MGRX font format
 *
 * FEdit is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FEdit is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FEdit. If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <grgui.h>
#include "fedit.h"

static void paint_dot(DotMatrix *dm, int x, int y, int blit);

DotMatrix *DotMatrixCreate(int width, int height, int dotsize, int dotfill)
{
    DotMatrix *dm;
    
    dm = (DotMatrix*)malloc(sizeof(DotMatrix));
    if (dm == NULL) return NULL;
    
    dm->dot = (char *)calloc(width*height, sizeof(char));
    if (dm->dot == NULL) {
        free(dm);
        return NULL;
    }
    
    dm->width = width;
    dm->height = height;
    dm->dotsize = dotsize;
    dm->dotfill = dotfill;
    dm->dotselec = 0;
    dm->hline = -1;
    dm->vline = -1;
    dm->xorg = 0;
    dm->yorg = 0;
    dm->fillcolor = GrBlack();
    dm->dotoncolor = GrWhite();
    dm->dotoffcolor = GrBlack();
    dm->dotselcolor = GrWhite();
    dm->hlinecolor = GrWhite();
    dm->vlinecolor = GrWhite();
    dm->p = NULL;
    
    return dm;
}

void DotMatrixSetPos(DotMatrix *dm, int xorg, int yorg)
{
    dm->xorg = xorg;
    dm->yorg = yorg;
}

void DotMatrixSetHLine(DotMatrix *dm, int hline)
{
    if (hline <= dm->height) dm->hline = hline;
}

void DotMatrixSetVLine(DotMatrix *dm, int vline)
{
    if (vline <= dm->width) dm->vline = vline;
}

void DotMatrixIncrVLine(DotMatrix *dm, int dir)
{
    int vline;
    
    vline = dm->vline + dir;
    if (vline <= dm->width && vline > 0)
        dm->vline = vline;
}

void DotMatrixSetColors(DotMatrix *dm, GrColor fill, GrColor doton,
                        GrColor dotoff, GrColor dotsel)
{
    dm->fillcolor = fill;
    dm->dotoncolor = doton;
    dm->dotoffcolor = dotoff;
    dm->dotselcolor = dotsel;
}

void DotMatrixSetLineColors(DotMatrix *dm, GrColor hline, GrColor vline)
{
    dm->hlinecolor = hline;
    dm->vlinecolor = vline;
}

void DotMatrixSetAllDots(DotMatrix *dm, int onoff)
{
    int i, max;

    max = dm->height * dm->width;
    for (i=0; i<max; i++) {
        dm->dot[i] = onoff;
    }
}

void DotMatrixSetDot(DotMatrix *dm, int x, int y, int onoff)
{
    int ind;
    
    if (x < 0 || x >= dm->width) return;
    if (y < 0 || y >= dm->height) return;
    
    ind = dm->width * y + x;
    dm->dot[ind] = onoff;
}

int DotMatrixSwtichDot(DotMatrix *dm, int x, int y)
{
    int ind;
    
    if (x < 0 || x >= dm->width) return 0;
    if (y < 0 || y >= dm->height) return 0;
    
    ind = dm->width * y + x;
    dm->dot[ind] = dm->dot[ind] ? 0 : 1;
    return dm->dot[ind];
}

int DotMatrixGetDot(DotMatrix *dm, int x, int y)
{
    int ind;
    
    if (x < 0 || x >= dm->width) return 0;
    if (y < 0 || y >= dm->height) return 0;
    
    ind = dm->width * y + x;
    return dm->dot[ind];
}

int DotMatrixCalcDotByGCoord(DotMatrix *dm, int gx, int gy, int *x, int *y, int m)
{
    int mx, my;

    gx -= dm->xorg + dm->dotfill;
    gy -= dm->yorg + dm->dotfill;
    if (gx < 0 || gy < 0) return 0;
    
    *x = gx / (dm->dotsize + dm->dotfill);
    mx = gx % (dm->dotsize + dm->dotfill);
    *y = gy / (dm->dotsize + dm->dotfill);
    my = gy % (dm->dotsize + dm->dotfill);
    //printf("%d %d %d %d\n", gx, gy, *x, *y);
    
    if (*x < 0 || *x >= dm->width) return 0;
    if (*y < 0 || *y >= dm->height) return 0;
    if (!m && (mx >= dm->dotsize || my >= dm->dotsize)) return 0;
    
    return 1;
}

void DotMatrixMoveX(DotMatrix *dm, int dir)
{
    int i, j, ind1, ind2;

    if (dir != -1 && dir != 1) return;
    
    if (dir == -1) {
        for (i=0; i<dm->height; i++) {
            for (j=1; j<dm->width; j++) {
                ind1 = dm->width * i + (j-1);
                ind2 = dm->width * i + j;
                dm->dot[ind1] = dm->dot[ind2];
            }
            ind1 = (dm->width * (i + 1)) - 1;
            dm->dot[ind1] = 0;
        }
    } else {
        for (i=0; i<dm->height; i++) {
            for (j=dm->width-1; j>0; j--) {
                ind1 = dm->width * i + (j-1);
                ind2 = dm->width * i + j;
                dm->dot[ind2] = dm->dot[ind1];
            }
            ind1 = dm->width * i;
            dm->dot[ind1] = 0;
        }
    }
}

void DotMatrixMoveY(DotMatrix *dm, int dir)
{
    int i, j, ind1, ind2;

    if (dir != -1 && dir != 1) return;
    
    if (dir == 1) {
        for (j=0; j<dm->width; j++) {
            for (i=1; i<dm->height; i++) {
                ind1 = dm->width * (i-1) + j;
                ind2 = dm->width * i + j;
                dm->dot[ind1] = dm->dot[ind2];
            }
            ind1 = (dm->width * (dm->height-1)) + j;
            dm->dot[ind1] = 0;
        }
    } else {
        for (j=0; j<dm->width; j++) {
            for (i=dm->height-1; i>0; i--) {
                ind1 = dm->width * (i-1) + j;
                ind2 = dm->width * i + j;
                dm->dot[ind2] = dm->dot[ind1];
            }
            ind1 = j;
            dm->dot[ind1] = 0;
        }
    }
}

void DotMatrixPaint(DotMatrix *dm)
{
    int width, height, i, j;
    int xdot, ydot;

    width = (dm->dotsize + dm->dotfill) * dm->width + dm->dotfill;
    height = (dm->dotsize + dm->dotfill) * dm->height + dm->dotfill;

    GrFilledBox(dm->xorg, dm->yorg, dm->xorg+width-1,
                dm->yorg+height-1, dm->fillcolor);
    for (i=0; i<dm->height; i++) {
        for (j=0; j<dm->width; j++) {
            paint_dot(dm, j, i, 0);
        }
    }

    if (dm->hline >= 0) {
        ydot = dm->yorg + dm->hline * (dm->dotsize + dm->dotfill);
        GrHLine(dm->xorg, dm->xorg+width-1, ydot, dm->hlinecolor);
        if (dm->dotfill > 1)
            GrHLine(dm->xorg, dm->xorg+width-1, ydot+1, dm->hlinecolor);
    }
    if (dm->vline >= 0) {
        xdot = dm->xorg + dm->vline * (dm->dotsize + dm->dotfill);
        GrVLine(xdot, dm->yorg, dm->yorg+height-1, dm->vlinecolor);
        if (dm->dotfill > 1)
            GrVLine(xdot+1, dm->yorg, dm->yorg+height-1, dm->vlinecolor);
    }

    GUIDBCurCtxBltRectToScreen(dm->xorg, dm->yorg, dm->xorg+width-1,
                               dm->yorg+height-1);
}

void DotMatrixRePaintDots(DotMatrix *dm)
{
    int width, height, i, j;

    width = (dm->dotsize + dm->dotfill) * dm->width + dm->dotfill;
    height = (dm->dotsize + dm->dotfill) * dm->height + dm->dotfill;

    for (i=0; i<dm->height; i++) {
        for (j=0; j<dm->width; j++) {
            paint_dot(dm, j, i, 0);
        }
    }

    GUIDBCurCtxBltRectToScreen(dm->xorg, dm->yorg,dm->xorg+width-1,
                               dm->yorg+height-1);
}

void DotMatrixRePaintLines(DotMatrix *dm, int show)
{
    int width, height;
    GrColor color;
    int xdot, ydot;

    width = (dm->dotsize + dm->dotfill) * dm->width + dm->dotfill;
    height = (dm->dotsize + dm->dotfill) * dm->height + dm->dotfill;

    if (dm->hline >= 0) {
        color = show ? dm->hlinecolor : dm->fillcolor;
        ydot = dm->yorg + dm->hline * (dm->dotsize + dm->dotfill);
        GrHLine(dm->xorg, dm->xorg+width-1, ydot, color);
        if (dm->dotfill > 1)
            GrHLine(dm->xorg, dm->xorg+width-1, ydot+1, color);
        GUIDBCurCtxBltRectToScreen(dm->xorg, ydot, dm->xorg+width-1, ydot+1);
    }

    if (dm->vline >= 0) {
        color = show ? dm->vlinecolor : dm->fillcolor;
        xdot = dm->xorg + dm->vline * (dm->dotsize + dm->dotfill);
        GrVLine(xdot, dm->yorg, dm->yorg+height-1, color);
        if (dm->dotfill > 1)
            GrVLine(xdot+1, dm->yorg, dm->yorg+height-1, color);
        GUIDBCurCtxBltRectToScreen(xdot, dm->yorg, xdot+1, dm->yorg+height-1);
    }
}

void DotMatrixPaintDot(DotMatrix *dm, int x, int y)
{
    if (x < 0 || x >= dm->width) return;
    if (y < 0 || y >= dm->height) return;

    paint_dot(dm, x, y, 1);
}

void DotMatrixSetSelect(DotMatrix *dm, int x, int y, int paint)
{
    int newselec, oldselec, ox, oy;

    if (x < 0 || x >= dm->width) return;
    if (y < 0 || y >= dm->height) return;

    newselec = dm->width * y + x;

    if (dm->dotselec != newselec) {
        oldselec = dm->dotselec;
        dm->dotselec = newselec;
        if (paint) {
            oy = (oldselec / dm->width);
            ox = (oldselec % dm->width);
            paint_dot(dm, ox, oy, 1);
        }
    }

    if (paint) { // paint newselect ever
        paint_dot(dm, x, y, 1);
    }
}

void DotMatrixMoveRelSelect(DotMatrix *dm, int row, int col, int paint)
{
    int numdot, newselec, oldselec, x, y;

    newselec = dm->dotselec + (row * dm->width) + col;
    if (newselec < 0) return;
    numdot = dm->width * dm->height;
    if (newselec >= numdot) return;

    if (dm->dotselec != newselec) {
        oldselec = dm->dotselec;
        dm->dotselec = newselec;
        if (paint) {
            y = (oldselec / dm->width);
            x = (oldselec % dm->width);
            paint_dot(dm, x, y, 1);
            y = (newselec / dm->width);
            x = (newselec % dm->width);
            paint_dot(dm, x, y, 1);
        }
    }
}

int DotMatrixSwitchSelect(DotMatrix *dm, int paint)
{
    int x, y, ret;

    y = (dm->dotselec / dm->width);
    x = (dm->dotselec % dm->width);
    ret = DotMatrixSwtichDot(dm, x, y);

    if (paint) {
        paint_dot(dm, x, y, 1);
    }

    return ret;
}

void DotMatrixReflectV(DotMatrix *dm)
{
    char aux[dm->width];
    int nrowsr = dm->height / 2;
    int i, j, r;

    for (i=0; i<nrowsr; i++) {
        r = dm->height - 1 - i;
        for (j=0; j<dm->width; j++)
            aux[j] = dm->dot[dm->width*i+j];
        for (j=0; j<dm->width; j++)
            dm->dot[dm->width*i+j] = dm->dot[dm->width*r+j];
        for (j=0; j<dm->width; j++)
            dm->dot[dm->width*r+j] = aux[j];
    }
}

void DotMatrixReflectH(DotMatrix *dm)
{
    char aux[dm->height];
    int ncolsr = dm->width / 2;
    int i, j, r;

    for (i=0; i<ncolsr; i++) {
        r = dm->width - 1 - i;
        for (j=0; j<dm->height; j++)
            aux[j] = dm->dot[dm->width*j+i];
        for (j=0; j<dm->height; j++)
            dm->dot[dm->width*j+i] = dm->dot[dm->width*j+r];
        for (j=0; j<dm->height; j++)
            dm->dot[dm->width*j+r] = aux[j];
    }
}

static void rec_fill(DotMatrix *dm, int x, int y, int bit)
{
    int ind;

    if (x < 0 || x >= dm->width) return;
    if (y < 0 || y >= dm->height) return;

    ind = dm->width * y + x;
    if (dm->dot[ind] == bit) return;
    dm->dot[ind] = bit;
    rec_fill(dm, x+1, y, bit);
    rec_fill(dm, x, y+1, bit);
    rec_fill(dm, x-1, y, bit);
    rec_fill(dm, x, y-1, bit);
}

void DotMatrixFill(DotMatrix *dm, int x, int y)
{
    int ind, bit;

    if (x < 0 || x >= dm->width) return;
    if (y < 0 || y >= dm->height) return;

    ind = dm->width * y + x;
    bit = dm->dot[ind] ? 0 : 1;
    rec_fill(dm, x, y, bit);
}

void DotMatrixDestroy(DotMatrix *dm)
{
    free(dm->dot);
    free(dm);
}

static void paint_dot(DotMatrix *dm, int x, int y, int blit)
{
    int ind;
    GrColor color;
    int xdot, ydot;
    
    ind = dm->width * y + x;
    color = dm->dot[ind] ? dm->dotoncolor : dm->dotoffcolor;
    xdot = dm->xorg + dm->dotfill + x * (dm->dotsize + dm->dotfill);
    ydot = dm->yorg + dm->dotfill + y * (dm->dotsize + dm->dotfill);
    GrFilledBox(xdot, ydot, xdot+dm->dotsize-1, ydot+dm->dotsize-1, color);
    if (ind == dm->dotselec) {
        //color = dm->dot[ind] ? dm->hlinecolor : dm->vlinecolor;
        //color = dm->dot[ind] ? dm->dotoffcolor : dm->dotoncolor;
        color = dm->dot[ind] ? dm->dotoffcolor : dm->dotselcolor;
        GrBox(xdot+1, ydot+1, xdot+dm->dotsize-2, ydot+dm->dotsize-2, color);
    }
    if (blit) {
        GUIDBCurCtxBltRectToScreen(xdot, ydot, xdot+dm->dotsize-1,
                                   ydot+dm->dotsize-1);
    }
}
