/*
 * fedit.c --- FEdit main source
 * 
 * Copyright (c) 2019 Mariano Alvarez Fernandez (malfer@telefonica.net)
 *
 * This file is part of FEdit, a visual font editor for the MGRX font format
 *
 * FEdit is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FEdit is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FEdit. If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <grgui.h>
#include <mgrxcolr.h>
#include "fedit.h"
#include "hershey.h"

#define COMMAND_LOADFONT      1
#define COMMAND_LOADHEFONT    2
#define COMMAND_SAVEFONT      3
#define COMMAND_APPENDFONT    4
#define COMMAND_EDITPRO       5
#define COMMAND_RSZBOX        6
#define COMMAND_RESIZE        7
#define COMMAND_SETFIXED      8
#define COMMAND_SETPROP       9
#define COMMAND_EDITGLYPH    10
#define COMMAND_ADDGLYPH     11
#define COMMAND_PASTEGLYPH   12
#define COMMAND_DELGLYPH     13
#define COMMAND_RLOADGLYPH   14
#define COMMAND_DELRANGE     20
#define COMMAND_RLOADRANGE   21
#define COMMAND_COPYRANGE    22
#define COMMAND_DMLEFT       30
#define COMMAND_DMRIGHT      31
#define COMMAND_DMUP         32
#define COMMAND_DMDOWN       33
#define COMMAND_DMWLEFT      34
#define COMMAND_DMWRIGHT     35
#define COMMAND_DMREFLECTV   36
#define COMMAND_DMREFLECTH   37
#define COMMAND_ABOUT        90
#define COMMAND_EXIT        100

#define ID_MENUFONT           1
#define ID_MENUGLYPH          2
#define ID_MENUDM             3
#define ID_MENUHELP           4
#define ID_MENUGLYPHF        11

#define IDT1 1
#define IDT2 2
#define IDT3 3
#define IDT4 4
#define IDT5 5

typedef struct {
    GUITile *gt;
    GrColor fg;
    GrColor bg;
} UserData;

static int usedoublebuffer = 0;
static int globalw = 1200;
static int globalh = 800;
static int globalbpp = 8;
static int globalmaxglyphs = 1024;
static char *glbfirstfname = NULL;
static char defaultfirstfname[] = "pc8x16";
static char glbinputfname[MAXLEN_FNAME+1] = "";
static char glboutputfname[MAXLEN_FNAME+1] = "output.fnt";
static GrFont *glbfont;
static GrTextOption glbgrt;
static DotMatrix *glbdm;
static FontWork *glbfwk;
static UserData ud1, ud2, ud3, ud4, ud5;

static void setup(int argc, char **argv);
static void terminate(void);
static void disaster(char *s);
static void setup_menus(void);
static void setup_tiles(void);
static void setup_glbfwk(void);
static void setup_glbdm(void);
static void load_first_glbfont(void);
static void load_glbfont(void);
static void load_hershey_glbfont(void);
static void adjust_menu_options(void);
static void save_glbfwkfont(void);
static void append_glbfwkfont(void);
static void edit_properties(void);
static void uninit_range(void);
static void copy_range(void);
static void reload_range(void);
static void resize_box(void);
static void resize_font(void);
static void convert_fixed_font(void);
static void convert_prop_font(void);
//static void convert_prop_font2(void);
static void about(void);
void paint_tile1(void *data);
void paint_tile2(void *data);
void paint_tile3(void *data);
void paint_tile4(void *data);
void paint_tile5(void *data);
int process_tile2_event(void *data, GrEvent *ev);
static void run_tile3_context_menu(void);
int process_tile3_event(void *data, GrEvent *ev);
//int load_char_in_dm(DotMatrix *dm, GrFont *f, unsigned int glyph);
int load_fwkglyph_in_dm(DotMatrix *dm, FontWork *fwk, int glyph, int add);
int save_dm_in_fwkglyph(DotMatrix *dm, FontWork *fwk, int glyph);

int main(int argc, char **argv)
{
    GrEvent ev;
    int ret;

    setup(argc, argv);
    setup_menus();
    load_first_glbfont();
    setup_tiles();
    setup_glbfwk();
    setup_glbdm();
    load_fwkglyph_in_dm(glbdm, glbfwk, 32, 0);
    adjust_menu_options();

    GUITilePaint(IDT1);
    GUITilePaint(IDT2);
    GUITilePaint(IDT3);
    GUITilePaint(IDT4);
    GUITilePaint(IDT5);

    while(1) {
        GrEventRead(&ev);
        if (ev.type == GREV_COMMAND) {
            if (ev.p1 == COMMAND_LOADFONT) {
                load_glbfont();
                continue;
            } else if (ev.p1 == COMMAND_LOADHEFONT) {
                load_hershey_glbfont();
                continue;
            } else if (ev.p1 == COMMAND_SAVEFONT) {
                save_glbfwkfont();
                continue;
            } else if (ev.p1 == COMMAND_APPENDFONT) {
                append_glbfwkfont();
                continue;
            } else if (ev.p1 == COMMAND_EDITPRO) {
                edit_properties();
                continue;
            } else if (ev.p1 == COMMAND_RSZBOX) {
                resize_box();
                continue;
            } else if (ev.p1 == COMMAND_RESIZE) {
                resize_font();
                continue;
            } else if (ev.p1 == COMMAND_SETFIXED) {
                convert_fixed_font();
                continue;
            } else if (ev.p1 == COMMAND_SETPROP) {
                convert_prop_font();
                continue;
            } else if ((ev.p1 == COMMAND_EDITGLYPH) ||
                       (ev.p1 == COMMAND_ADDGLYPH) ||
                       (ev.p1 == COMMAND_PASTEGLYPH) ||
                       (ev.p1 == COMMAND_DELGLYPH) ||
                       (ev.p1 == COMMAND_RLOADGLYPH)) {
                GUITilesProcessTileEvent(&ev, IDT2);
                continue;
            } else if ((ev.p1 == COMMAND_DMLEFT) ||
                       (ev.p1 == COMMAND_DMRIGHT) ||
                       (ev.p1 == COMMAND_DMUP) ||
                       (ev.p1 == COMMAND_DMDOWN) ||
                       (ev.p1 == COMMAND_DMWLEFT) ||
                       (ev.p1 == COMMAND_DMWRIGHT) ||
                       (ev.p1 == COMMAND_DMREFLECTV) ||
                       (ev.p1 == COMMAND_DMREFLECTH)) {
                GUITilesProcessTileEvent(&ev, IDT3);
                continue;
            } else if (ev.p1 == COMMAND_DELRANGE) {
                uninit_range();
                continue;
            } else if (ev.p1 == COMMAND_COPYRANGE) {
                copy_range();
                continue;
            } else if (ev.p1 == COMMAND_RLOADRANGE) {
                reload_range();
                continue;
            } else if (ev.p1 == COMMAND_ABOUT) {
                about();
                continue;
            } else if (ev.p1 == COMMAND_EXIT) {
                break;
            }
        } else if (ev.type == GREV_WMEND) {
            break;
        }
        ret = GUITilesProcessEvent(&ev);
        if (ret == -1) break;
    }
    
    terminate();
    return 0;
}

/**************************/

static void setup(int argc, char **argv)
{
    static char *wintitle = "FEdit 1.2";
    int ind, error = 0;

    ind = 1;
    while (1) {
        if (ind >= argc) break;
        if (strcmp(argv[ind], "--gm") == 0) {
            if (ind+3 >= argc) {
                error = 1;
                break;
            }
            globalw = atoi(argv[ind+1]);
            globalh = atoi(argv[ind+2]);
            globalbpp = atoi(argv[ind+3]);
            ind += 4;
        } else if (strcmp(argv[ind], "--mx") == 0) {
            if (ind+1 >= argc) {
                error = 1;
                break;
            }
            globalmaxglyphs = atoi(argv[ind+1]);
            ind += 2;
        } else if (strcmp(argv[ind], "--db") == 0) {
            usedoublebuffer = 1;
            ind++;
        } else if (strncmp(argv[ind], "--", 2) == 0) {
            error = 1;
            break;
        } else {
            glbfirstfname = argv[ind];
            ind++;
        }
    }

    if (error) {
        printf("Usage: fedit [--gm width height bpp] [--mx maxglyphs]"
               " [--db] [fontname]\n");
        exit(1);
    }

    if (glbfirstfname != NULL) {
        strncpy(glbinputfname, glbfirstfname, MAXLEN_FNAME);
        glbinputfname[MAXLEN_FNAME] = '\0';
        if (strstr(glbfirstfname, ".fnt") != NULL) {
            strncpy(glboutputfname, glbfirstfname, MAXLEN_FNAME);
            glboutputfname[MAXLEN_FNAME] = '\0';
        }
    } else {
        glbfirstfname = defaultfirstfname;
    }

    if (!GrSetMode(GR_width_height_bpp_graphics, globalw, globalh, globalbpp)) {
        printf("Error setting the graphics mode\n");
        exit(1);
    }

    GrEventGenWMEnd(GR_GEN_WMEND_YES);
    GrSetUserEncoding(GRENC_UTF_8);
    GrGenWebColorTable();
    GrSetFontPath("./fonts/;./");
    GrSetDefaultFont(&GrFont_PX8x18);
    //GUIInit(1, 0);
    GUIInit(1, usedoublebuffer);
    //GrEventGenExpose(GR_GEN_EXPOSE_YES);
    GrSetWindowTitle(wintitle);

    GUIObjectsSetColors(WEBC_BLACK, WEBC_LIGHTGRAY, WEBC_DIMGRAY);
    GUIDialogsSetColors(WEBC_BLACK, WEBC_YELLOW, WEBC_STEELBLUE, WEBC_WHITE);
    GUIDialogsSetTitleFont(&GrFont_PX11x22);
    GUICDialogsSetColors(WEBC_GREEN, WEBC_WHITE);
}

/**************************/

static void terminate(void)
{
    GUITilesDestroyAll();
    DotMatrixDestroy(glbdm);
    FontWorkDestroy(glbfwk);
    GrUnloadFont(glbfont);
    GUIEnd();
    GrSetMode(GR_default_text);
}

/**************************/

static void disaster(char *s)
{
    void _GrCloseVideoDriver(void);

    char aux[81];

    GUIEnd();
    _GrCloseVideoDriver();
    printf("FEdit: Disaster %s\n", s);
    printf("press Return to continue\n");
    fgets(aux, 80, stdin);
    exit(1);
}

/**************************/

static void setup_menus(void)
{
    static GUIMenuItem itemsm1[14] = {
        {GUI_MI_OPER, 1, "&Load Font...", 'L', NULL, 0, COMMAND_LOADFONT, 0},
        {GUI_MI_OPER, 1, "&Save Font...", 'S', NULL, 0, COMMAND_SAVEFONT, 0},
        {GUI_MI_SEP, 1, "", 0, NULL, 0, 0, 0}, 
        {GUI_MI_OPER, 1, "Load Font from &Hershey data...", 'H', NULL, 0, COMMAND_LOADHEFONT, 0},
        {GUI_MI_OPER, 1, "&Append glyphs from Font...", 'A', NULL, 0, COMMAND_APPENDFONT, 0},
        {GUI_MI_SEP, 1, "", 0, NULL, 0, 0, 0}, 
        {GUI_MI_OPER, 1, "&Edit properties...", 'E', NULL, 0, COMMAND_EDITPRO, 0},
        {GUI_MI_OPER, 1, "Convert to &fixed Font...", 'F', NULL, 0, COMMAND_SETFIXED, 0},
        {GUI_MI_OPER, 1, "Convert to &prop. Font...", 'P', NULL, 0, COMMAND_SETPROP, 0},
        {GUI_MI_SEP, 1, "", 0, NULL, 0, 0, 0}, 
        {GUI_MI_OPER, 1, "&Resize Font box...", 'R', NULL, 0, COMMAND_RSZBOX, 0},
        {GUI_MI_OPER, 1, "Resi&ze Font...", 'Z', NULL, 0, COMMAND_RESIZE, 0},
        {GUI_MI_SEP, 1, "", 0, NULL, 0, 0, 0}, 
        {GUI_MI_OPER, 1, "E&xit", 'X', "Alt+X", GrKey_Alt_X, COMMAND_EXIT, 0}};
    static GUIMenu menu1 = {ID_MENUFONT, 14, 0, itemsm1};

    static GUIMenuItem itemsm2[9] = {
        {GUI_MI_OPER, 1, "&Edit glyph in DotArea", 'E', "Alt+E", GrKey_Alt_E, COMMAND_EDITGLYPH, 0},
        {GUI_MI_OPER, 1, "&Add glyph to DotArea", 'A', "Alt+A", GrKey_Alt_A, COMMAND_ADDGLYPH, 0},
        {GUI_MI_OPER, 1, "&Paste glyph from DotArea", 'P', "Alt+P", GrKey_Alt_P, COMMAND_PASTEGLYPH, 0},
        {GUI_MI_OPER, 1, "&Uninit glyph", 'U', "Alt+U", GrKey_Alt_U, COMMAND_DELGLYPH, 0},
        {GUI_MI_OPER, 1, "&Reload glyph", 'R', "Alt+R", GrKey_Alt_R, COMMAND_RLOADGLYPH, 0},
        {GUI_MI_SEP, 1, "", 0, NULL, 0, 0, 0}, 
        {GUI_MI_OPER, 1, "U&ninit glyph range...", 'N', NULL, 0, COMMAND_DELRANGE, 0},
        {GUI_MI_OPER, 1, "&Copy glyph range...", 'C', NULL, 0, COMMAND_COPYRANGE, 0},
        {GUI_MI_OPER, 1, "Re&load glyph range...", 'L', NULL, 0, COMMAND_RLOADRANGE, 0}};
    static GUIMenu menu2 = {ID_MENUGLYPH, 9, 0, itemsm2};

    static GUIMenuItem itemsm3[10] = {
        {GUI_MI_OPER, 1, "Move glyph &left", 'L', "Ctrl+Left", GrKey_Control_Left, COMMAND_DMLEFT, 0},
        {GUI_MI_OPER, 1, "Move glyph &right", 'R', "Ctrl+Right", GrKey_Control_Right, COMMAND_DMRIGHT, 0},
        {GUI_MI_OPER, 1, "Move glyph &up", 'U', "Ctrl+Up", GrKey_Control_Up, COMMAND_DMUP, 0},
        {GUI_MI_OPER, 1, "Move glyph &down", 'D', "Ctrl+Down", GrKey_Control_Down, COMMAND_DMDOWN, 0},
        {GUI_MI_SEP, 1, "", 0, NULL, 0, 0, 0}, 
        {GUI_MI_OPER, 1, "Move width bar le&ft", 'F', "Shift+Left", GrKey_Shift_Left, COMMAND_DMWLEFT, 0},
        {GUI_MI_OPER, 1, "Move width bar ri&ght", 'G', "Shift+Right", GrKey_Shift_Right, COMMAND_DMWRIGHT, 0},
        {GUI_MI_SEP, 1, "", 0, NULL, 0, 0, 0}, 
        {GUI_MI_OPER, 1, "Reflect &vertically", 'V', NULL, 0, COMMAND_DMREFLECTV, 0},
        {GUI_MI_OPER, 1, "Reflect &horizontally", 'H', NULL, 0, COMMAND_DMREFLECTH, 0}};
    static GUIMenu menu3 = {ID_MENUDM, 10, 0, itemsm3};

    static GUIMenuItem itemsm4[1] = {
        {GUI_MI_OPER, 1, "&About FEdit", 'A', NULL, 0, COMMAND_ABOUT, 0}};
    static GUIMenu menu4 = {ID_MENUHELP, 1, 0, itemsm4};

    static GUIMenuItem itemsm11[5] = {
        {GUI_MI_OPER, 1, "&Edit glyph in DotArea", 'E', NULL, 0, COMMAND_EDITGLYPH, 0},
        {GUI_MI_OPER, 1, "&Add glyph to DotArea", 'A', NULL, 0, COMMAND_ADDGLYPH, 0},
        {GUI_MI_OPER, 1, "&Paste glyph from DotArea", 'P', NULL, 0, COMMAND_PASTEGLYPH, 0},
        {GUI_MI_OPER, 1, "&Uninit glyph", 'U', NULL, 0, COMMAND_DELGLYPH, 0},
        {GUI_MI_OPER, 1, "&Reload glyph", 'R', NULL, 0, COMMAND_RLOADGLYPH, 0}};
    static GUIMenu menu11 = {ID_MENUGLYPHF, 5, 0, itemsm11};

    static GUIMenuBarItem mbitems[4] = {
        {"&Font", 1, GrKey_Alt_F, ID_MENUFONT}, 
        {"&Glyph", 1, GrKey_Alt_G, ID_MENUGLYPH},
        {"&DotArea", 1, GrKey_Alt_D, ID_MENUDM},
        {"&Help", 1, GrKey_Alt_H, ID_MENUHELP}};
    static GUIMenuBar menubar = {4 ,0, mbitems};

    GUIMenusSetColors(WEBC_WHITE, WEBC_DARKRED, WEBC_BLACK,
                      WEBC_YELLOW, WEBC_LIGHTGRAY, WEBC_DARKGRAY);
    GUIMenusSetFont(&GrFont_PX11x22);

    GUIMenuRegister(&menu1);
    GUIMenuRegister(&menu2);
    GUIMenuRegister(&menu3);
    GUIMenuRegister(&menu4);
    GUIMenuRegister(&menu11);
    GUIMenuBarSet(&menubar);
    GUIMenuBarShow();
}

/**************************/

static void setup_tiles(void)
{
    GUITile *gt1, *gt2, *gt3, *gt4, *gt5;
    int yorg;

    yorg = GUIMenuBarGetHeight();

    gt1 = GUITileCreate(IDT1, GUI_TT_STATICBORDER, 0, yorg,
                        GrSizeX(), 40);
    gt2 = GUITileCreate(IDT2, GUI_TT_ACTIVEBWVSCB, 0, yorg+40,
                        GrSizeX()/2, GrSizeY()-(yorg+80));
    gt3 = GUITileCreate(IDT3, GUI_TT_ACTIVEBORDER, GrSizeX()/2, yorg+40,
                        GrSizeX()/2, GrSizeY()-(yorg+80));
    gt4 = GUITileCreate(IDT4, GUI_TT_BORDERLESS, 0,
                        GrSizeY()-40, GrSizeX()/2, 40);
    gt5 = GUITileCreate(IDT5, GUI_TT_BORDERLESS, GrSizeX()/2,
                        GrSizeY()-40, GrSizeX()/2, 40);
    if (gt1 == NULL || gt2 == NULL || gt3 == NULL || gt4 == NULL || gt5 == NULL) {
        disaster("creating tiles");
    }
    
    GUIPanelSetClCallBacks(gt1->p, paint_tile1, NULL);
    GUIPanelSetUserData(gt1->p, (void *)&ud1);
    ud1.gt = gt1;
    ud1.fg = GrWhite();
    ud1.bg = WEBC_STEELBLUE;
    GUIPanelSetClCallBacks(gt2->p, paint_tile2, process_tile2_event);
    GUIPanelSetUserData(gt2->p, (void *)&ud2);
    ud2.gt = gt2;
    ud2.fg = GrWhite();
    ud2.bg = WEBC_GREEN;
    GUIPanelSetClCallBacks(gt3->p, paint_tile3, process_tile3_event);
    GUIPanelSetUserData(gt3->p, (void *)&ud3);
    ud3.gt = gt3;
    ud3.fg = GrWhite();
    ud3.bg = WEBC_GREEN;
    GUIPanelSetClCallBacks(gt4->p, paint_tile4, NULL);
    GUIPanelSetUserData(gt4->p, (void *)&ud4);
    ud4.gt = gt4;
    ud4.fg = GrWhite();
    ud4.bg = WEBC_DIMGRAY;
    GUIPanelSetClCallBacks(gt5->p, paint_tile5, NULL);
    GUIPanelSetUserData(gt5->p, (void *)&ud5);
    ud5.gt = gt5;
    ud5.fg = GrWhite();
    ud5.bg = WEBC_DIMGRAY;

    GUITileRegister(gt1);
    GUITileRegister(gt2);
    GUITileRegister(gt3);
    GUITileRegister(gt4);
    GUITileRegister(gt5);

    GUIScrollbarsSetColors(WEBC_LIGHTGRAY, WEBC_DIMGRAY);
    GUITilesSetColors(WEBC_BLACK, WEBC_LIGHTGRAY, WEBC_YELLOW);
}

/**************************/

static void setup_glbfwk(void)
{
    GUITile *t;
    
    t = GUITileGetPointer(IDT2);
    if (t == NULL) disaster("getting tile pointer");

    glbfwk = FontWorkCreate(glbfont, globalmaxglyphs);
    if (glbfwk == NULL) disaster("creating fontwork");

    FontWorkSetDims(glbfwk, t);
    FontWorkSetColors(glbfwk, WEBC_DARKGREEN, WEBC_WHITE, WEBC_YELLOW, WEBC_RED);
}

/**************************/

static void setup_glbdm(void)
{
    GUITile *t;
    int tw, th;
    int glbdmw, glbdmh;
    int sw, sh, sf;
    int dotsize, fillsize;
    int x, y;

    t = GUITileGetPointer(IDT3);
    if (t == NULL) disaster("getting tile pointer");
    tw = t->p->cl->gc_xmax + 1;
    th = t->p->cl->gc_ymax + 1;

    glbdmw = glbfwk->maxwidth;
    glbdmh = glbfwk->h.height;
    sw = tw / (glbdmw + 1);
    sh = th / (glbdmh + 1);
    sf = (sw > sh) ? sh : sw;
    if (sf < 5) sf = 5;
    if (sf > 30) sf = 30;
    fillsize = (sf > 8) ? 2 : 1;
    dotsize = sf - fillsize;

    glbdm = DotMatrixCreate(glbdmw, glbdmh, dotsize, fillsize);
    if (glbdm == NULL) disaster("creating dot matrix");
    DotMatrixSetHLine(glbdm, glbfwk->h.baseline+1);
    if (glbfwk->h.proportional)
        DotMatrixSetVLine(glbdm, 1);
    else
        DotMatrixSetVLine(glbdm, glbfwk->maxwidth);
    DotMatrixSetColors(glbdm, WEBC_DIMGRAY, WEBC_WHITE, WEBC_BLACK, WEBC_YELLOW);
    DotMatrixSetLineColors(glbdm, WEBC_LIME, WEBC_CYAN);
    
    x = (tw - (sf * glbdmw) - fillsize) / 2;
    y = (th - (sf * glbdmh) - fillsize) / 2;
    DotMatrixSetPos(glbdm, x, y);
    glbdm->p = t->p;
    //printf("%d %d %d %d\n", glbdmw, glbdmh, tw, th);
    //printf("%d %d %d %d\n", dotsize, fillsize, x, y);
}

/**************************/

static void load_first_glbfont(void)
{
    glbfont = GrLoadFont(glbfirstfname);
    if (glbfont == NULL) disaster("loading glbfont");

    glbgrt.txo_font = glbfont;
    glbgrt.txo_fgcolor = WEBC_WHITE;
    glbgrt.txo_bgcolor = WEBC_GREEN;
    glbgrt.txo_chrtype = GrGetChrtypeForUserEncoding();
    glbgrt.txo_direct = GR_TEXT_RIGHT;
    glbgrt.txo_xalign = GR_ALIGN_LEFT;
    glbgrt.txo_yalign = GR_ALIGN_TOP;
}

/**************************/

static void load_glbfont(void)
{
    int ret;
    GrFont *f;

    FLoadDlgCreate();
    FLoadDlgSetData(glbinputfname);
    ret = FLoadDlgRun();
    if (ret < 0) return;
    FLoadDlgGetData(glbinputfname);

    f = GrLoadFont(glbinputfname);
    if (f == NULL) {
        ShowError2cat("Error loading font ", glbinputfname);
        return;
    }

    DotMatrixDestroy(glbdm);
    FontWorkDestroy(glbfwk);
    GrUnloadFont(glbfont);
    glbfont = f;
    glbgrt.txo_font = glbfont;

    setup_glbfwk();
    setup_glbdm();
    load_fwkglyph_in_dm(glbdm, glbfwk, 32, 0);

    adjust_menu_options();

    GUITilePaint(IDT1);
    GUITilePaint(IDT2);
    GUITilePaint(IDT3);
    GUITilePaint(IDT4);
    GUITilePaint(IDT5);
}

/**************************/

static void load_hershey_glbfont(void)
{
    static char fname[MAXLEN_FNAME+1];
    int maxglyph=globalmaxglyphs, fdglyph=0, thick=1;
    double zoom=1.0;
    int ret, err;
    char s[81], *p;
    HersheyFont *hf;
    GrFont *f;
    int newwidth, newheight;

    FLoheDlgCreate();
    FLoheDlgSetData(fname, maxglyph, fdglyph, zoom, thick);
    ret = FLoheDlgRun();
    if (ret < 0) return;
    FLoheDlgGetData(fname, &maxglyph, &fdglyph, &zoom, &thick);

    hf = LoadHersheyFont(fname, fdglyph, maxglyph, &err);
    if (hf == NULL) {
        sprintf(s, "Error %d loading Hershey font", err);
        ShowError2cat(s, fname);
        return;
    }
    GetHersheyFontBox(hf, zoom, &newwidth, &newheight);
    if ((newwidth < 4) || (newwidth > MAXBOX_W) || 
        (newheight < 4) || (newheight > MAXBOX_H)) {
        ShowErrorReqBoxInvalid(newwidth, newheight);
        UnloadHersheyFont(hf);
        return;
    }
    p = strrchr(fname, '/');
    if (p == NULL) p = fname;
    else p++;
    f = GenGrxHersheyFont(hf, p, "hershey", zoom, thick);
    UnloadHersheyFont(hf);
    if (f == NULL) {
        ShowError2cat("Error generating MGRX font", "from Hershey font");
        return;
    }

    DotMatrixDestroy(glbdm);
    FontWorkDestroy(glbfwk);
    GrUnloadFont(glbfont);
    glbfont = f;
    glbgrt.txo_font = glbfont;

    setup_glbfwk();
    setup_glbdm();
    load_fwkglyph_in_dm(glbdm, glbfwk, 32, 0);

    adjust_menu_options();

    GUITilePaint(IDT1);
    GUITilePaint(IDT2);
    GUITilePaint(IDT3);
    GUITilePaint(IDT4);
    GUITilePaint(IDT5);
}

/**************************/

static void adjust_menu_options(void)
{
    if (glbfwk->h.proportional) {
        GUIMenuSetEnable(ID_MENUDM, GUI_MI_OPER, COMMAND_DMWLEFT, 1);
        GUIMenuSetEnable(ID_MENUDM, GUI_MI_OPER, COMMAND_DMWRIGHT, 1);
        GUIMenuSetEnable(ID_MENUFONT, GUI_MI_OPER, COMMAND_SETFIXED, 1);
        GUIMenuSetEnable(ID_MENUFONT, GUI_MI_OPER, COMMAND_SETPROP, 0);
    } else {
        GUIMenuSetEnable(ID_MENUDM, GUI_MI_OPER, COMMAND_DMWLEFT, 0);
        GUIMenuSetEnable(ID_MENUDM, GUI_MI_OPER, COMMAND_DMWRIGHT, 0);
        GUIMenuSetEnable(ID_MENUFONT, GUI_MI_OPER, COMMAND_SETFIXED, 0);
        GUIMenuSetEnable(ID_MENUFONT, GUI_MI_OPER, COMMAND_SETPROP, 1);
    }
}

/**************************/

static void save_glbfwkfont(void)
{
    GrFont *f;
    int fglyph, lglyph;
    int ret;
    FILE *ftry;

    fglyph = glbfwk->h.minchar;
    lglyph = glbfwk->h.minchar + glbfwk->h.numchars - 1;

    FSaveDlgCreate();
    FSaveDlgSetData(glboutputfname, fglyph, lglyph);

    while (1) {
        ret = FSaveDlgRun();
        if (ret < 0) return;
        FSaveDlgGetData(glboutputfname, &fglyph, &lglyph);
        if (!FontWorkCheckValidRange(glbfwk, fglyph, lglyph)) {
            ShowError("No valid range, try again");
            continue;
        }
        if (FontWorkCheckUninitedInRange(glbfwk, fglyph, lglyph)) {
            ShowError("Uninited glyphs in range, try again");
            continue;
        }
        break;
    }

    ftry = fopen(glboutputfname, "r");
    if (ftry != NULL) {
        fclose(ftry);
        ret = WarnContinue(glboutputfname, "already exist");
        if (ret != 1) return;
    }
    
    glbfwk->h.minchar = fglyph;
    glbfwk->h.numchars = lglyph - glbfwk->h.minchar + 1;

    f = FontWorkGenFont(glbfwk);
    if (f == NULL) {
        ShowError("Internal error generating new font");
        return;
    }
        
    ret = GrDumpGrxFont(f, glboutputfname);
    if (!ret) {
        ShowError2cat("Error saving font file ", glboutputfname);
        return;
    }
    
    GrUnloadFont(glbfont);
    glbfont = f;
    glbgrt.txo_font = glbfont;
    GUITilePaint(IDT1);
}

/**************************/

static void append_glbfwkfont(void)
{
    static char fontname[MAXLEN_FNAME+1];
    int fsglyph, lsglyph, fdglyph, xpivot, ypivot;
    int ret;
    GrFont *f;

    fsglyph = 0;
    lsglyph = 255;
    fdglyph = glbfwk->h.minchar + glbfwk->h.numchars;
    xpivot = 0;
    ypivot = 0;

    FApenDlgCreate();
    FApenDlgSetData(fontname, fsglyph, lsglyph, fdglyph, xpivot, ypivot);

    while (1) {
        ret = FApenDlgRun();
        if (ret < 0) return;
        FApenDlgGetData(fontname, &fsglyph, &lsglyph, &fdglyph, &xpivot, &ypivot);
        if (!FontWorkCheckValidRange(glbfwk, fdglyph, fdglyph+lsglyph-fsglyph+1)) {
            ShowError("No valid dest. range, try again");
            continue;
        }
        break;
    }

    f = GrLoadFont(fontname);
    if (f == NULL) {
        ShowError2cat("Error loading font ", fontname);
        return;
    }

    if ((fsglyph < f->h.minchar) || (lsglyph >= f->h.minchar + f->h.numchars)) {
        ShowError2cat("No valid source range for font ", fontname);
        return;
    }

    ret = FontWorkAppend(glbfwk, f, fsglyph, lsglyph, fdglyph, xpivot, ypivot);
    GrUnloadFont(f);
    if (ret == 0) {
        ShowError("Internal error appending font");
        return;
    }

    GUITilePaint(IDT2);
}

/**************************/

static void edit_properties(void)
{
    int ret, oldbaseline;

    oldbaseline = glbfwk->h.baseline;
    FPropDlgCreate();
    FPropDlgSetData(glbfwk);
    ret = FPropDlgRun();
    if (ret < 0) return;

    FPropDlgGetData(glbfwk);
    GUITilePaint(IDT1);

    if (oldbaseline != glbfwk->h.baseline) {
        DotMatrixSetHLine(glbdm, glbfwk->h.baseline+1);
        GUITilePaint(IDT3);
    }
}

/**************************/

static void uninit_range(void)
{
    static int fglyph = 0;
    static int lglyph = 0;
    int ret, i;

    URangDlgCreate();
    URangDlgSetData(fglyph, lglyph);

    while (1) {
        ret = URangDlgRun();
        if (ret < 0) return;
        URangDlgGetData(&fglyph, &lglyph);
        if (!FontWorkCheckValidRange(glbfwk, fglyph, lglyph)) {
            ShowError("No valid range, try again");
            continue;
        }
        break;
    }

    for (i=fglyph; i<=lglyph; i++) {
        glbfwk->gly[i].inited = 0;
    }
    
    GUITilePaint(IDT2);
}

/**************************/

static void copy_range(void)
{
    static int fsglyph = 0;
    static int lsglyph = 0;
    static int fdglyph = 0;
    int ret, i, ldglyph;

    CRangDlgCreate();
    CRangDlgSetData(fsglyph, lsglyph, fdglyph);

    while (1) {
        ret = CRangDlgRun();
        if (ret < 0) return;
        CRangDlgGetData(&fsglyph, &lsglyph, &fdglyph);
        ldglyph = fdglyph + (lsglyph - fsglyph);
        if ((lsglyph >= fsglyph) && (fsglyph >= 0) &&
            (lsglyph < glbfwk->maxglyphs) &&
            (ldglyph >= fdglyph) && (fdglyph >= 0) &&
            (ldglyph < glbfwk->maxglyphs) &&
            (fsglyph != fdglyph))
            break;
        ShowError("No valid range, try again");
    }

    if (fdglyph > fsglyph) {
        for (i=lsglyph; i>=fsglyph; i--) {
            FontWorkCopyGlyph(glbfwk, i, i+(fdglyph-fsglyph));
        }
    } else {
        for (i=fsglyph; i<=lsglyph; i++) {
            FontWorkCopyGlyph(glbfwk, i, i-(fsglyph-fdglyph));
        }
    }
    
    GUITilePaint(IDT2);
}

/**************************/

static void reload_range(void)
{
    static int fglyph = 0;
    static int lglyph = 0;
    int ret, i;

    RRangDlgCreate();
    RRangDlgSetData(fglyph, lglyph);

    while (1) {
        ret = RRangDlgRun();
        if (ret < 0) return;
        RRangDlgGetData(&fglyph, &lglyph);
        if (!FontWorkCheckValidRange(glbfwk, fglyph, lglyph)) {
            ShowError("No valid range, try again");
            continue;
        }
        break;
    }

    for (i=fglyph; i<=lglyph; i++) {
        FontWorkLoadGlyph(glbfwk, glbfont, i);
    }
    
    GUITilePaint(IDT2);
}

/**************************/

static void resize_box(void)
{
    static int top = 0;
    static int left = 0;
    static int right = 0;
    static int bottom = 0;
    FontWork *fwk2;
    GrFont *f2;
    int reqwidth, reqheight;
    int ret;

    if (FontWorkCheckUninitedInFontRange(glbfwk)) {
        ShowError2("There are uninited glyphs in font range",
                   "plese init them and try again");
        return;
    }

    RsBoxDlgCreate();
    RsBoxDlgSetData(top, left, right, bottom);

    while (1) {
        ret = RsBoxDlgRun();
        if (ret < 0) return;
        RsBoxDlgGetData(&top, &left, &right, &bottom);
        reqwidth = glbfwk->maxwidth + left + right;
        reqheight = glbfwk->h.height + top + bottom;
        // we impose a minimal box of 4x4 an maximum of 100x100
        if ((reqwidth >= 4) && (reqwidth <= MAXBOX_W) && 
            (reqheight >= 4) && (reqheight <= MAXBOX_H)) break;
        ShowErrorReqBoxInvalid(reqwidth, reqheight);
    }

    fwk2 = FontWorkResizeBox(glbfwk, top, left, right, bottom);
    if (fwk2 == NULL) {
        ShowError("Internal error generating new font");
        return;
    }

    f2 = FontWorkGenFont(fwk2);
    if (f2 == NULL) {
        ShowError("Internal error generating new font");
        FontWorkDestroy(fwk2);
        return;
    }
    GrUnloadFont(glbfont);
    glbfont = f2;

    FontWorkSetSelect(fwk2, glbfwk->glyphselec, 0);
    DotMatrixDestroy(glbdm);
    FontWorkDestroy(glbfwk);
    glbfwk = fwk2;
    glbgrt.txo_font = glbfont;
    setup_glbdm();
    load_fwkglyph_in_dm(glbdm, glbfwk, glbfwk->glyphselec, 0);

    GUITilePaint(IDT1);
    GUITilePaint(IDT2);
    GUITilePaint(IDT3);
    GUITilePaint(IDT4);
    GUITilePaint(IDT5);
}

/**************************/

static void resize_font(void)
{
    GrFont *f, *f2;
    int newwidth, newheight, fglyph, lglyph;
    int ret, glyphselec;

    //newwidth = glbfwk->maxwidth;
    FontWorkRecalMeanWidth(glbfwk);
    newwidth = glbfwk->h.width;
    newheight = glbfwk->h.height;
    fglyph = glbfwk->h.minchar;
    lglyph = glbfwk->h.minchar + glbfwk->h.numchars - 1;

    RSizeDlgCreate();
    RSizeDlgSetData(newwidth, newheight, fglyph, lglyph);

    while (1) {
        ret = RSizeDlgRun();
        if (ret < 0) return;
        RSizeDlgGetData(&newwidth, &newheight, &fglyph, &lglyph);
        if (!FontWorkCheckValidRange(glbfwk, fglyph, lglyph)) {
            ShowError("No valid range, try again");
            continue;
        }
        if (FontWorkCheckUninitedInRange(glbfwk, fglyph, lglyph)) {
            ShowError("Uninited glyphs in range, try again");
            continue;
        }
        // we impose a minimal box of 4x4 an maximum of 100x100
        if ((newwidth < 4) || (newwidth > MAXBOX_W) || 
            (newheight < 4) || (newheight > MAXBOX_H)) {
            ShowErrorReqBoxInvalid(newwidth, newheight);
            continue;
        }
        if (FontWorkCheckInitedOutOfRange(glbfwk, fglyph, lglyph)) {
            ret = WarnContinue("There are inited glyphs out of the range",
                               "which will be lost after resizing");
            if (ret != 1) continue;
        }
        break;
    }

    glbfwk->h.minchar = fglyph;
    glbfwk->h.numchars = lglyph - glbfwk->h.minchar + 1;
    glyphselec = glbfwk->glyphselec;

    f = FontWorkGenFont(glbfwk);
    if (f) {
        f2 = GrBuildConvertedFont(f, GR_FONTCVT_RESIZE, newwidth, newheight, 0, 0);
    }
    if ((f == NULL) || (f2 == NULL)) {
        ShowError("Internal error generating new font");
        return;
    }
        
    GrUnloadFont(glbfont);
    GrUnloadFont(f);
    glbfont = f2;
    glbgrt.txo_font = glbfont;
    
    DotMatrixDestroy(glbdm);
    FontWorkDestroy(glbfwk);
    setup_glbfwk();
    setup_glbdm();
    FontWorkSetSelect(glbfwk, glyphselec, 0);
    load_fwkglyph_in_dm(glbdm, glbfwk, glbfwk->glyphselec, 0);

    //adjust_menu_options();

    GUITilePaint(IDT1);
    GUITilePaint(IDT2);
    GUITilePaint(IDT3);
    GUITilePaint(IDT4);
    GUITilePaint(IDT5);
}

/**************************/

static void convert_fixed_font(void)
{
    FontWork *fwk2;
    GrFont *f2;

    if (!glbfwk->h.proportional) return;

    if (FontWorkCheckUninitedInFontRange(glbfwk)) {
        ShowError2("There are uninited glyphs in font range",
                   "plese init them and try again");
        return;
    }

    if (ConfirmContinue("Change font to fixed width", "") != 1) return;
    
    fwk2 = FontWorkConvToFixed(glbfwk);
    if (fwk2 == NULL) {
        ShowError("Internal error generating new font");
        return;
    }

    f2 = FontWorkGenFont(fwk2);
    if (f2 == NULL) {
        ShowError("Internal error generating new font");
        FontWorkDestroy(fwk2);
        return;
    }
    GrUnloadFont(glbfont);
    glbfont = f2;

    FontWorkSetSelect(fwk2, glbfwk->glyphselec, 0);
    DotMatrixDestroy(glbdm);
    FontWorkDestroy(glbfwk);
    glbfwk = fwk2;
    glbgrt.txo_font = glbfont;
    setup_glbdm();
    load_fwkglyph_in_dm(glbdm, glbfwk, glbfwk->glyphselec, 0);

    adjust_menu_options();

    GUITilePaint(IDT1);
    GUITilePaint(IDT2);
    GUITilePaint(IDT3);
    GUITilePaint(IDT4);
    GUITilePaint(IDT5);
}

/**************************/

static void convert_prop_font(void)
{
    GrFont *f2;

    if (glbfwk->h.proportional) return;

    if (FontWorkCheckUninitedInFontRange(glbfwk)) {
        ShowError2("There are uninited glyphs in font range",
                   "plese init them and try again");
        return;
    }

    if (ConfirmContinue("Change font to proportional width", "") != 1) return;

    glbfwk->h.proportional = 1;

    f2 = FontWorkGenFont(glbfwk);
    if (f2 == NULL) {
        glbfwk->h.proportional = 0;
        ShowError("Internal error generating new font");
        return;
    }
    GrUnloadFont(glbfont);
    glbfont = f2;

    glbgrt.txo_font = glbfont;

    adjust_menu_options();

    GUITilePaint(IDT1);
}

/**************************/
/*
static void convert_prop_font2(void)
{
    GrFont *f, *f2;
    int glyphselec;

    if (glbfwk->h.proportional) return;

    if (FontWorkCheckUninitedInFontRange(glbfwk)) {
        ShowError2("There are uninited glyphs in font range",
                   "plese init them and try again");
        return;
    }

    if (ConfirmContinue("Change font to proportional width", "") != 1) return;

    glyphselec = glbfwk->glyphselec;

    f = FontWorkGenFont(glbfwk);
    if (f) {
        f2 = GrBuildConvertedFont(f, GR_FONTCVT_PROPORTION, 0, 0, 0, 0);
    }
    if ((f == NULL) || (f2 == NULL)) {
        ShowError("Internal error generating new font");
        return;
    }
        
    GrUnloadFont(glbfont);
    GrUnloadFont(f);
    glbfont = f2;
    glbgrt.txo_font = glbfont;
    
    DotMatrixDestroy(glbdm);
    FontWorkDestroy(glbfwk);
    setup_glbfwk();
    setup_glbdm();
    FontWorkSetSelect(glbfwk, glyphselec, 0);
    load_fwkglyph_in_dm(glbdm, glbfwk, glbfwk->glyphselec, 0);

    adjust_menu_options();

    GUITilePaint(IDT1);
    GUITilePaint(IDT2);
    GUITilePaint(IDT3);
    GUITilePaint(IDT4);
    GUITilePaint(IDT5);
}
*/
/**************************/

static void about(void)
{
    char *atext[5] = {"FEdit 1.2, (C)2019-2020 Mariano Alvarez",
                      "A small and nice Font Editor",
                      "for the MRGX font format",
                      "FEdit is Free Software,",
                      "see readme.txt file for details"};

    GUICDialogInfo("About FEdit", (void **)atext, 5, "Ok");
}

/**************************/

void paint_tile1(void *data)
{
    UserData *ud;
    char s[121+MAXLEN_FNAME], *p;

    ud = (UserData *)data;
    GrClearContext(ud->bg);
    p = s;
    p += sprintf(p, "Font name: %s", glbfwk->h.name);
    if (glbfwk->h.proportional) {
        p += sprintf(p, " (%d-%d)x%d proportional", glbfont->minwidth,
                     glbfwk->maxwidth, glbfwk->h.height);
    } else {
        p += sprintf(p, " %dx%d fixed", glbfwk->h.width, glbfwk->h.height);
    }
    p += sprintf(p, "  Min/Max glyphs: %d/%d ", glbfwk->h.minchar,
                 glbfwk->h.minchar + glbfwk->h.numchars - 1);
    p += sprintf(p, "  Encoding: %d (%s) ", glbfwk->h.encoding,
                  GrStrFontEncoding(glbfwk->h.encoding));
    p += sprintf(p, "  File: %s ", glboutputfname);
    GrTextXY(10, 12, s, ud->fg, ud->bg);
}

/**************************/

void paint_tile2(void *data)
{
    UserData *ud;

    ud = (UserData *)data;
    GrClearContext(ud->bg);
    FontWorkPaint(glbfwk);
}

/**************************/

void paint_tile3(void *data)
{
    UserData *ud;

    ud = (UserData *)data;
    GrClearContext(ud->bg);
    DotMatrixPaint(glbdm);
}

/**************************/

void paint_tile4(void *data)
{
    UserData *ud;
    char s[121], *p;
    long rc;

    ud = (UserData *)data;
    GrClearContext(ud->bg);
    p = s;
    p += sprintf(p, "Glyph: %d (%04X)  ", glbfwk->glyphselec, glbfwk->glyphselec);
    if (glbfwk->gly[glbfwk->glyphselec].inited) {
        p += sprintf(p, "width: %d  ", glbfwk->gly[glbfwk->glyphselec].width);
        if (glbfont->h.encoding != 0) {
            rc = GrFontCharReverseRecode(glbfont, glbfwk->glyphselec, GR_UCS2_TEXT);
            if (rc != 0x3f && glbfwk->glyphselec != 0x3f)
                p += sprintf(p, "Unicode: 0x%04X", (int)rc);
        }
    } else {
        p += sprintf(p, "No inited  ");
    }

    GrTextXY(10, 12, s, ud->fg, ud->bg);
}

/**************************/

void paint_tile5(void *data)
{
    UserData *ud;
    char s[121];
    int x, y;

    ud = (UserData *)data;
    GrClearContext(ud->bg);

    y = (glbdm->dotselec / glbdm->width);
    x = (glbdm->dotselec % glbdm->width);
    sprintf(s, "X:%d  Y:%d  width: %d", x, y, glbdm->vline);
    
    GrTextXY(10, 12, s, ud->fg, ud->bg);
}

/**************************/

int process_tile2_event(void *data, GrEvent *ev)
{
    //UserData *ud;
    int glyph;

    //ud = (UserData *)data;

    if (ev->type == GREV_COMMAND) {
        if (ev->p1 == COMMAND_EDITGLYPH) {
            if (load_fwkglyph_in_dm(glbdm, glbfwk, glbfwk->glyphselec, 0)) {
                GUITilePaint(IDT3);
                GUITilePaint(IDT5);
            }
            return 1;
        } else if (ev->p1 == COMMAND_ADDGLYPH) {
            if (load_fwkglyph_in_dm(glbdm, glbfwk, glbfwk->glyphselec, 1)) {
                GUITilePaint(IDT3);
                GUITilePaint(IDT5);
            }
            return 1;
        } else if (ev->p1 == COMMAND_PASTEGLYPH) {
            save_dm_in_fwkglyph(glbdm, glbfwk, glbfwk->glyphselec);
            FontWorkPaintGlyph(glbfwk, glbfwk->glyphselec, 1, 1);
            GUITilePaint(IDT4);
            return 1;
        } else if (ev->p1 == COMMAND_DELGLYPH) {
            glbfwk->gly[glbfwk->glyphselec].inited = 0;
            FontWorkPaintGlyph(glbfwk, glbfwk->glyphselec, 1, 1);
            GUITilePaint(IDT4);
            return 1;
        } else if (ev->p1 == COMMAND_RLOADGLYPH) {
            glbfwk->gly[glbfwk->glyphselec].inited = 0;
            FontWorkLoadGlyph(glbfwk, glbfont, glbfwk->glyphselec);
            FontWorkPaintGlyph(glbfwk, glbfwk->glyphselec, 1, 1);
            GUITilePaint(IDT4);
            return 1;
        }
    }

    if (ev->type == GREV_SCBVERT) {
        FontWorkScrollSelect(glbfwk, ev->p1, 1);
        GUITilePaint(IDT4);
        return 1;
    }

    if (ev->type == GREV_KEY) {
        if (ev->p2 == GRKEY_KEYCODE) {
            if (ev->p1 == GrKey_Right) {
                FontWorkMoveSelect(glbfwk, 0, 1, 1);
                GUITilePaint(IDT4);
                return 1;
            } else if (ev->p1 == GrKey_Left) {
                FontWorkMoveSelect(glbfwk, 0, -1, 1);
                GUITilePaint(IDT4);
                return 1;
            } else if (ev->p1 == GrKey_Up) {
                FontWorkMoveSelect(glbfwk, -1, 0, 1);
                GUITilePaint(IDT4);
                return 1;
            } else if (ev->p1 == GrKey_Down) {
                FontWorkMoveSelect(glbfwk, 1, 0, 1);
                GUITilePaint(IDT4);
                return 1;
            } else if (ev->p1 == GrKey_PageUp) {
                //FontWorkScrollSelect(glbfwk,
                //    glbfwk->firstline-(glbfwk->screenlines-1), 1);
                FontWorkMovePagSelect(glbfwk, -1, 1);
                GUITilePaint(IDT4);
                return 1;
            } else if (ev->p1 == GrKey_PageDown) {
                //FontWorkScrollSelect(glbfwk,
                //    glbfwk->firstline+(glbfwk->screenlines-1), 1);
                FontWorkMovePagSelect(glbfwk, 1, 1);
                GUITilePaint(IDT4);
                return 1;
            }
        } else {
            if (ev->p1 == '\r') {
                run_tile3_context_menu();
                return 1;
            }
        }
    }
    if (ev->type == GREV_MOUSE) {
        if (ev->p1 == GRMOUSE_LB_PRESSED) {
            if (!FontWorkCalcGlyphByGCoord(glbfwk, ev->p2, ev->p3, &glyph))
                return 1;
            FontWorkSetSelect(glbfwk, glyph, 1);
            GUITilePaint(IDT4);
            return 1;
        } else if (ev->p1 == GRMOUSE_RB_PRESSED) {
            if (!FontWorkCalcGlyphByGCoord(glbfwk, ev->p2, ev->p3, &glyph))
                return 1;
            FontWorkSetSelect(glbfwk, glyph, 1);
            GUITilePaint(IDT4);
            run_tile3_context_menu();
            return 1;
        } else if (ev->p1 == GRMOUSE_B4_RELEASED) {
            FontWorkMoveSelect(glbfwk, -1, 0, 1);
            GUITilePaint(IDT4);
            return 1;
        } else if (ev->p1 == GRMOUSE_B5_RELEASED) {
            FontWorkMoveSelect(glbfwk, 1, 0, 1);
            GUITilePaint(IDT4);
            return 1;
        }
    }

    return 0;
}

/**************************/

static void run_tile3_context_menu(void)
{
    int x, y;

    if (glbfwk->gly[glbfwk->glyphselec].inited) {
        GUIMenuSetEnable(ID_MENUGLYPHF, GUI_MI_OPER, COMMAND_EDITGLYPH, 1);
        GUIMenuSetEnable(ID_MENUGLYPHF, GUI_MI_OPER, COMMAND_ADDGLYPH, 1);
        GUIMenuSetEnable(ID_MENUGLYPHF, GUI_MI_OPER, COMMAND_DELGLYPH, 1);
    } else {
        GUIMenuSetEnable(ID_MENUGLYPHF, GUI_MI_OPER, COMMAND_EDITGLYPH, 0);
        GUIMenuSetEnable(ID_MENUGLYPHF, GUI_MI_OPER, COMMAND_ADDGLYPH, 0);
        GUIMenuSetEnable(ID_MENUGLYPHF, GUI_MI_OPER, COMMAND_DELGLYPH, 0);
    }
    FontWorkGetRealSelectedXYPos(glbfwk, &x, &y);
    // this is a hack by now
    int maxy = y+(22*5+4);
    if (y+(22*5+4) > GrScreenY()) {
        y -= maxy - GrScreenY();
    }
    // end of the hack
    GUIMenuRun(ID_MENUGLYPHF, x, y, 0);
}

/**************************/

int process_tile3_event(void *data, GrEvent *ev)
{
    //UserData *ud;
    int xd, yd;
    static int last = 0;
    int glyph;

    //ud = (UserData *)data;

    if (ev->type == GREV_COMMAND) {
        if (ev->p1 == COMMAND_DMRIGHT) {
            DotMatrixMoveX(glbdm, 1);
            DotMatrixRePaintDots(glbdm);
            return 1;
        } else if (ev->p1 == COMMAND_DMLEFT) {
            DotMatrixMoveX(glbdm, -1);
            DotMatrixRePaintDots(glbdm);
            return 1;
        } else if (ev->p1 == COMMAND_DMUP) {
            DotMatrixMoveY(glbdm, 1);
            DotMatrixRePaintDots(glbdm);
            return 1;
        } else if (ev->p1 == COMMAND_DMDOWN) {
            DotMatrixMoveY(glbdm, -1);
            DotMatrixRePaintDots(glbdm);
            return 1;
        } else if (ev->p1 == COMMAND_DMWRIGHT) {
            if (!glbfwk->h.proportional) return 0;
            DotMatrixRePaintLines(glbdm, 0);
            DotMatrixIncrVLine(glbdm, 1);
            DotMatrixRePaintLines(glbdm, 1);
            GUITilePaint(IDT5);
            return 1;
        } else if (ev->p1 == COMMAND_DMWLEFT) {
            if (!glbfwk->h.proportional) return 0;
            DotMatrixRePaintLines(glbdm, 0);
            DotMatrixIncrVLine(glbdm, -1);
            DotMatrixRePaintLines(glbdm, 1);
            GUITilePaint(IDT5);
            return 1;
        } else if (ev->p1 == COMMAND_DMREFLECTV) {
            DotMatrixReflectV(glbdm);
            DotMatrixRePaintDots(glbdm);
            return 1;
        } else if (ev->p1 == COMMAND_DMREFLECTH) {
            DotMatrixReflectH(glbdm);
            DotMatrixRePaintDots(glbdm);
            return 1;
        }
    }

    if (ev->type == GREV_KEY) {
        if (ev->p2 != GRKEY_KEYCODE) {
            if (ev->p1 == '\r') {
                DotMatrixSwitchSelect(glbdm, 1);
            } else {
                glyph = GrFontCharRecode(glbfont, ev->p1,
                                         GrGetChrtypeForUserEncoding());
                if (load_fwkglyph_in_dm(glbdm, glbfwk, glyph, 0)) {
                    DotMatrixPaint(glbdm);
                    GUITilePaint(IDT5);
                }
                return 1;
            }
        } else {
            if (ev->p1 == GrKey_Right) {
                DotMatrixMoveRelSelect(glbdm, 0, 1, 1);
                GUITilePaint(IDT5);
                return 1;
            } else if (ev->p1 == GrKey_Left) {
                DotMatrixMoveRelSelect(glbdm, 0, -1, 1);
                GUITilePaint(IDT5);
                return 1;
            } else if (ev->p1 == GrKey_Up) {
                DotMatrixMoveRelSelect(glbdm, -1, 0, 1);
                GUITilePaint(IDT5);
                return 1;
            } else if (ev->p1 == GrKey_Down) {
                DotMatrixMoveRelSelect(glbdm, 1, 0, 1);
                GUITilePaint(IDT5);
                return 1;
            }
       }
    }
    if (ev->type == GREV_MOUSE) {
        if (ev->p1 == GRMOUSE_LB_PRESSED) {
            if (DotMatrixCalcDotByGCoord(glbdm, ev->p2, ev->p3, &xd, &yd, 0)) {
                last = DotMatrixSwtichDot(glbdm, xd, yd);
                DotMatrixSetSelect(glbdm, xd, yd, 1);
                GUITilePaint(IDT5);
            }
            return 1;
        } else if (ev->p1 == GRMOUSE_MB_PRESSED) {
            if (DotMatrixCalcDotByGCoord(glbdm, ev->p2, ev->p3, &xd, &yd, 0)) {
                DotMatrixSetSelect(glbdm, xd, yd, 0);
                DotMatrixFill(glbdm, xd, yd);
                DotMatrixRePaintDots(glbdm);
                GUITilePaint(IDT5);
            }
            return 1;
        }
    }
    if (ev->type == GREV_MMOVE) {
        if ((ev->p1 & GRMOUSE_LB_STATUS)) {
            if (DotMatrixCalcDotByGCoord(glbdm, ev->p2, ev->p3, &xd, &yd, 1)) {
                DotMatrixSetDot(glbdm, xd, yd, last);
                DotMatrixSetSelect(glbdm, xd, yd, 1);
                GUITilePaint(IDT5);
            }
            return 1;
        }
    }
    return 0;
}

/**************************/
/*
int load_char_in_dm(DotMatrix *dm, GrFont *f, unsigned int glyph)
{
    int fcw, fch, fcbrs;
    char *bmp, *p;
    int i, j;
    int byte, bit;

    if (!GrFontCharPresent(f, glyph)) return 0;
    fcw = GrFontCharWidth(f, glyph);
    fch = GrFontCharHeight(f, glyph);
    fcbrs = GrFontCharBmpRowSize(f, glyph);
    //fcbms = GrFontCharBitmapSize(f, glyph);
    bmp = GrFontCharBitmap(f, glyph);
    
    DotMatrixSetAllDots(dm, 0);
    DotMatrixSetHLine(dm, f->h.baseline+1);
    DotMatrixSetVLine(dm, fcw);

    for (i=0; i<fch; i++) {
        p = bmp + (fcbrs*i);
        for (j=0; j<fcw; j++) {
            byte = j / 8;
            bit = 0x80 >> (j % 8);
            if (p[byte] & bit) {
                DotMatrixSetDot(dm, j, i, 1);
            }
        }
    }

    return 1;
}
*/
/**************************/

int load_fwkglyph_in_dm(DotMatrix *dm, FontWork *fwk, int glyph, int add)
{
    char *p;
    int i, j;
    int byte, bit;

    if (!fwk->gly[glyph].inited) return 0;
    if (glyph < 0 || glyph >= fwk->maxglyphs) return 0;

    if (!add) {
        DotMatrixSetAllDots(dm, 0);
        DotMatrixSetHLine(dm, fwk->h.baseline+1);
        DotMatrixSetVLine(dm, fwk->gly[glyph].width);
    }

    for (i=0; i<fwk->h.height; i++) {
        p = fwk->gly[glyph].bmp + (fwk->gly[glyph].bmprowsize * i);
        for (j=0; j<fwk->gly[glyph].width; j++) {
            byte = j / 8;
            bit = 0x80 >> (j % 8);
            if (p[byte] & bit)
                DotMatrixSetDot(dm, j, i, 1);
        }
    }

    return 1;
}

/**************************/

int save_dm_in_fwkglyph(DotMatrix *dm, FontWork *fwk, int glyph)
{
    char *p;
    int i, j;
    int byte, bit;

    if (glyph < 0 || glyph >= fwk->maxglyphs) return 0;
    if (dm->vline > fwk->maxwidth) return 0;

    fwk->gly[glyph].inited = 1;
    fwk->gly[glyph].width = dm->vline;
    fwk->gly[glyph].bmprowsize = (dm->vline + 7) / 8;
    memset(fwk->gly[glyph].bmp, 0, fwk->gly[glyph].bmprowsize * fwk->h.height);

    for (i=0; i<fwk->h.height; i++) {
        p = fwk->gly[glyph].bmp + (fwk->gly[glyph].bmprowsize * i);
        for (j=0; j<fwk->gly[glyph].width; j++) {
            byte = j / 8;
            bit = 0x80 >> (j % 8);
            if (DotMatrixGetDot(dm, j, i))
                p[byte] |= bit;
        }
    }

    return 1;
}
